function [disparity, lmatch] = marrPoggio(l, r, display)
% marrPoggio - Marr-Poggio Stereo Algorithm
%
% Usage: [disparity, lmatch] = marrPoggio(left, right)
% Runs the Marr-Poggio Stereo Algorithm.

% Validate input.
if nargin < 2,
    error('Usage: [disparity, lmatch] = marrPoggio(left, right)');
end

if nargin < 3
    display = [];
end

% Check that the images are equal in size, and binary.
rows = size(l, 1);
cols = size(l, 2);
if (size(r, 1) ~= rows) | (size(r, 2) ~= cols),
    error('The two images must have the same size.');
end;
if any(any((l == 0) + (l == 1) ~= ones(size(l)))) | any(any((r == 0) + (r == 1) ~= ones(size(r)))),
    error('The images must be binary.');
end;

% Algorithm parameters
%
% epsilon: Weight for the mismatch penalty
% exWidth: The width of the excitation region
% maxdisp: Maximum disparity
% maxiter: Maximum number of iterations
% mindisp: Minimum disparity
% sigma:   Threshold to be exceeded to flip a match on
% sigma1:  Same as sigma, but for first iteration
%
epsilon = 0.2;
exWidth = 5;
maxdisp = 3;
maxiter = 10;
mindisp = -3;
sigma   = 0;
sigma1  = -0.2;

% Count the number of possible disparities, given the limits.
ndisp = maxdisp - mindisp + 1;

% Initialize the lmatch array.
lmatch = zeros(rows, cols, ndisp);
n = 1;
for d = mindisp : maxdisp,
    ms = max(1, 1 - d);
    me = min(cols, cols - d);
    rs = max(1, 1 + d);
    re = min(cols, cols + d);
    ls = ms;
    le = me;
    lmatch(:, ms : me, n) = 1 - xor(l(:, ls : le), r(:, rs : re));
    n = n + 1;
end;

% Initialize other intermediate arrays.
mismatch   = zeros(size(lmatch));
excitation = zeros(size(lmatch));
inhibition = zeros(size(lmatch));
rmatch     = zeros(size(lmatch));

% Compute a convolution kernel for excitation calculations.
exKernel = ones(exWidth, 1);
exSize   = exWidth ^ 2 - 1;

% The mismatch variable flags matches between pixels with different values.
mismatch = 1 - lmatch;

% Iteratively apply the Marr-Poggio procedure.
for iter = 1 : maxiter,
    
    % Measure the amount of nearby pixels with the same disparity.
    n = 1;
    for d = mindisp : maxdisp,
        excitation(:, :, n) = conv2(conv2(lmatch(:, :, n), exKernel, 'same'), exKernel', 'same') - lmatch(:, :, n);
        n = n + 1;
    end;
    
    % Normalize the excitation metric so that a neighborhood without any matches produces the sigma=0 threshold.
    excitation = excitation / exSize;
    
    % Measure the amount of multiple matches for the same pixel in either image.  
    % Convert lmatch to rmatch.
    n = 1;
    for d = mindisp : maxdisp,
        rs = max(1, 1 + d);
        re = min(cols, cols + d);
        ls = max(1, 1 - d);
        le = min(cols, cols - d);
        rmatch(:, rs : re, n) = lmatch(:, ls : le, n);
        n = n + 1;
    end;
    
    % Count the multiple matches in the rows of both formats.  
    lMultiple(:, :, 1) = sum(lmatch, 3);
    rMultiple(:, :, 1) = sum(rmatch, 3);
    for i = 2 : ndisp,
        lMultiple(:, :, i) = lMultiple(:, :, 1);
        rMultiple(:, :, i) = rMultiple(:, :, 1);
    end;
    lMultiple = lMultiple - lmatch;
    rMultiple = rMultiple - rmatch;
    
    % Convert rMultiple back to the lmatch format.  
    n = 1;
    rMultiple2 = zeros(size(lmatch));
    for d = mindisp : maxdisp,
        rs = max(1, 1 + d);
        re = min(cols, cols + d);
        ls = max(1, 1 - d);
        le = min(cols, cols - d);
        rMultiple2(:, ls : le, n) = rMultiple(:, rs : re, n);
        n = n + 1;
    end;
    
    % Inhibition normalization
    inhibition = lMultiple + rMultiple2;
    inhibition = inhibition ./ (1 + inhibition);
    
    % In parallel, evaluate the support for each pixel.
    support = excitation - inhibition - epsilon * mismatch;
    
    % Update the lmatch array.
    oldmatch = lmatch;
    if iter == 1,
        lmatch = (support >= sigma1);
    else,
        lmatch = (support >= sigma);
    end;
    
    if ~isempty(display)
        rr = 1:20:rows;
        nn = length(rr);
        for k = 1:nn
            r = rr(k);
            subplot(nn, 1, k)
            imagesc(squeeze(lmatch(floor(rows/2), :, :))')
            axis equal
            axis image
            axis off
        end
        colormap gray
        figure(gcf)
        disp('Press any key to continue')
        pause
    end
    
    % Stop before maxiter iterations if not enough matches have changed.
    if sum(sum(sum(lmatch ~= oldmatch))) < max(5, rows * cols / 1000),
        break;
    end;
end;

% Reported disparities are the medians of disparities found at every pixel.
disp('Done with the matching; computing median disparities.')
disp('This may take a while because of a double for loop.')
disp('This would be much faster in C.')
disparity = zeros(rows, cols);
delta = (mindisp : maxdisp);
for i = 1 : rows,
    for j = 1 : cols,
        d = find(squeeze(lmatch(i, j, :)));
        if isempty(d),
            disparity(i, j) = NaN;
        else
            disparity(i, j) = median(delta(d));
        end;
    end;
end;
